 /*----------------------------------------------------------------------------------------------------+
 |                             National Institute of Standards and Technology                          |
 |                                        Biometric Clients Lab                                        |
 +-----------------------------------------------------------------------------------------------------+
  File author(s):
       Kevin Mangold (kevin.mangold@nist.gov)
 
 +-----------------------------------------------------------------------------------------------------+
 | NOTICE & DISCLAIMER                                                                                 |
 |                                                                                                     |
 | The research software provided on this web site ("software") is provided by NIST as a public        |
 | service. You may use, copy and distribute copies of the software in any medium, provided that you   |
 | keep intact this entire notice. You may improve, modify and create derivative works of the software |
 | or any portion of the software, and you may copy and distribute such modifications or works.        |
 | Modified works should carry a notice stating that you changed the software and should note the date |
 | and nature of any such change.  Please explicitly acknowledge the National Institute of Standards   |
 | and Technology as the source of the software.                                                       |
 |                                                                                                     |
 | The software is expressly provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED,  |
 | IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF      |
 | MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.  NIST        |
 | NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR         |
 | ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED.  NIST DOES NOT WARRANT OR MAKE ANY               |
 | REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED |
 | TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.                           |
 |                                                                                                     |
 | You are solely responsible for determining the appropriateness of using and distributing the        |
 | software and you assume all risks associated with its use, including but not limited to the risks   |
 | and costs of program errors, compliance with applicable laws, damage to or loss of data, programs   |
 | or equipment, and the unavailability or interruption of operation.  This software is not intended   |
 | to be used in any situation where a failure could cause risk of injury or damage to property.  The  |
 | software was developed by NIST employees.  NIST employee contributions are not subject to copyright |
 | protection within the United States.                                                                |
 |                                                                                                     |
 | Specific hardware and software products identified in this open source project were used in order   |
 | to perform technology transfer and collaboration. In no case does such identification imply         |
 | recommendation or endorsement by the National Institute of Standards and Technology, nor            |
 | does it imply that the products and equipment identified are necessarily the best available for the |
 | purpose.                                                                                            |
 +----------------------------------------------------------------------------------------------------*/

package gov.nist.itl.wsbd;

import static gov.nist.itl.wsbd.Logging.trace;
import gov.nist.itl.wsbd.concurrency.Operation;
import gov.nist.itl.wsbd.concurrency.OperationManager;
import gov.nist.itl.wsbd.concurrency.OperationPurpose;
import gov.nist.itl.wsbd.concurrency.ServiceLock;
import gov.nist.itl.wsbd.concurrency.SessionManager;
import gov.nist.itl.wsbd.dom.Dictionary;
import gov.nist.itl.wsbd.dom.Result;
import gov.nist.itl.wsbd.dom.Status;

import java.io.InputStream;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import javax.ws.rs.core.Response;

public abstract class SensorService implements ISensorService {
	protected static Map<Class<?>, ServiceLock> serviceLocks;
	
	public abstract Result getServiceInformation();	
	public abstract Result initialize(UUID sessionId) throws InterruptedException;	
	public abstract Result getConfiguration(UUID sessionId) throws InterruptedException;	
	public abstract Result setConfiguration(UUID sessionId, Dictionary configuration) throws InterruptedException;	
	public abstract Result capture(UUID sessionId) throws InterruptedException;	
	public abstract Result getDownloadInformation(UUID captureId) throws InterruptedException;
	public abstract Result thriftyDownload(UUID captureId, String maxSize) throws InterruptedException;
	public abstract Result download(UUID captureId) throws InterruptedException;
	public abstract InputStream stream();
	
	static {
//		serviceLocks = new HashMap<Class<?>, ServiceLock>();
	}
	
	public SensorService() {
		trace("SensorService()");
	}
	
	public static void reset() {
		if (serviceLocks != null) {
			serviceLocks.clear();
		}
	}
	
	public synchronized static ServiceLock getServiceLock(ISensorService service) {
		trace("SensorService.getServiceLock(ISensorService)");
		
		if (serviceLocks == null) {
			serviceLocks = new ConcurrentHashMap<Class<?>, ServiceLock>();
		}
		
		if (!serviceLocks.containsKey(service.getClass())) {
			serviceLocks.put(service.getClass(), new ServiceLock());
		}
		
		return serviceLocks.get(service.getClass());
	}
	
	public Result register() {
		trace("SensorService.register()");

		Result result = new Result();
		
		UUID sessionId = SessionManager.getInstance(this).registerSession();

		if (sessionId == null) {
			result.setStatus(Status.FAILURE);
		} else {
			result.setStatus(Status.SUCCESS);
			result.setSessionId(sessionId);
		}
		
		return result;
	}
	
	public Result unregister(UUID sessionId) {
		trace("SensorService.unregister(UUID)");
		
		Result result = new Result();

		if (true) {
			if (SessionManager.getInstance(this).unregisterSession(sessionId)) {
				result.setStatus(Status.SUCCESS);
			} else {
				result.setStatus(Status.FAILURE);
			}
		} else {
			result.setStatus(Status.FAILURE);
		}
		
		return result;
	}
	
	public Result lock(UUID sessionId) {
		trace("SensorService.lock(UUID)");
		
		Result result = new Result();
		
		ServiceLock serviceLock = getServiceLock(this);
		
		if (serviceLock.isLocked()) {
			if (serviceLock.holdsLock(sessionId)) {
				result.setStatus(Status.SUCCESS);
			} else {
				result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
			}
		} else {
			boolean locked = serviceLock.lock(sessionId);
			if (locked) {
				result.setStatus(Status.SUCCESS);
			} else {
				result.setStatus(Status.FAILURE);
			}
		}
		
		/*boolean locked = serviceLock.lock(sessionId);
		
		if (locked) {
			if (serviceLock.holdsLock(sessionId)) {
				result.setStatus(Status.SUCCESS);
			} else {
				result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
			}
		} else {
			result.setStatus(Status.LOCK_NOT_HELD);
		}*/
		
		return result;
	}
	
	public Result stealLock(UUID sessionId) {
		trace("SensorService.stealLock(UUID)");
		
		Result result = new Result();
		
		ServiceLock serviceLock = getServiceLock(this);
		
		boolean locked = serviceLock.stealLock(sessionId);
		
		if (locked) {
			result.setStatus(Status.SUCCESS);
		} else {
			result.setStatus(Status.FAILURE);
		}
		
		return result;
	}
	
	public Result unlock(UUID sessionId) {
		trace("SensorService.unlock(UUID)");
		
		Result result = new Result();
		
		ServiceLock serviceLock = getServiceLock(this);
		
		boolean unlocked = serviceLock.unlock(sessionId);
		
		if (unlocked) {
			result.setStatus(Status.SUCCESS);
		} else {
			result.setStatus(Status.FAILURE);
		}
		
		return result;
	}
	
	public Result cancel(UUID id) {
		trace("SensorService.cancel(UUID)");
		Result result = new Result();
		
		if (Utility.isValidId(id)) {
			if (getServiceLock(this).isLocked()) {
				if (getServiceLock(this).holdsLock(id)) {
					if (OperationManager.getManager().cancelCurrent()) {
						result.setStatus(Status.SUCCESS);
					} else {
						result.setStatus(Status.FAILURE);
					}
				} else {
					result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
				}
			} else {
				result.setStatus(Status.LOCK_NOT_HELD);
			}
		} else {
			result.setStatus(Status.INVALID_ID);
		}
		
		
		
		//success
		//failure
		//invalidId
		//lockNotHeld
		//lockHeldByAnother
		//badValue
		
		//getCurrentOperation().interrupt();
		
		return result;
	}
	
	@Override
	public boolean equals(Object service) {
		trace("SensorService.equals(Object)");
		
		if (service instanceof ISensorService) {
			if (service.getClass() == ISensorService.class) {
				if (service == this) {
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		} else {
			return false;
		}
	}
	
	@Override
	public int compareTo(ISensorService service) {
		trace("SensorService.compareTo(ISensorService)");
		
		if (this == service) {
			return 0;
		} else {
			return 1;
		}
	}

	@Override
	public Result IRegister() {
		trace("SensorService.IRegister()");

		Operation op = new Operation(OperationPurpose.REGISTER) {
			@Override
			public Result execute() {
				return register();
			}
		};

		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IUnregister(final UUID sessionId) {
		trace("SensorService.IUnregister(UUID)");
		
		Operation op = new Operation(OperationPurpose.UNREGISTER) {
			@Override
			public Result execute() {
				return unregister(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result ILock(final UUID sessionId) {
		trace("SensorService.ILock(UUID)");
		
		Operation op = new Operation(OperationPurpose.LOCK) {
			@Override
			public Result execute() {
				return lock(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IStealLock(final UUID sessionId) {
		trace("SensorService.IStealLock(UUID)");
		
		Operation op = new Operation(OperationPurpose.STEAL_LOCK) {
			@Override
			public Result execute() {
				return stealLock(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}
	
	@Override
	public Result IUnlock(final UUID sessionId) {
		trace("SensorService.IUnlock(UUID)");
		
		Operation op = new Operation(OperationPurpose.UNLOCK) {
			@Override
			public Result execute() {
				return unlock(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}
	
	@Override
	public Result ICancel(final UUID sessionId) {
		trace("SensorService.ICancel(UUID)");
		
		Operation op = new Operation(OperationPurpose.CANCEL) {
			@Override
			public Result execute() {
				return cancel(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IGetServiceInformation() {
		trace("SensorService.IGetServiceInformation()");
		
		Operation op = new Operation(OperationPurpose.GET_SERVICE_INFORMATION) {
			@Override
			public Result execute() {
				return getServiceInformation();
			}
		};

		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IInitialize(final UUID sessionId) throws InterruptedException {
		trace("SensorService.IInitialize(UUID)");
		
		Operation op = new Operation(OperationPurpose.INITIALIZE) {
			@Override
			public Result execute() throws InterruptedException {
				Result result = null;
				result = initialize(sessionId);
				
				return result;
			}
		};
		
		OperationManager.getManager().addOperation(op, true);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IGetConfiguration(final UUID sessionId) throws InterruptedException {
		trace("SensorService.IGetConfiguration(UUID)");
		
		Operation op = new Operation(OperationPurpose.GET_CONFIGURATION) {
			@Override
			public Result execute() throws InterruptedException {
				return getConfiguration(sessionId);
			}
		};
		
		OperationManager.getManager().addOperation(op, true);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result ISetConfiguration(final UUID sessionId, final Dictionary configuration) throws InterruptedException {
		trace("SensorService.ISetConfiguration(UUID, Directory)");
		
		Operation op = new Operation(OperationPurpose.SET_CONFIGURATION) {
			@Override
			public Result execute() throws InterruptedException {
				return setConfiguration(sessionId, configuration);
			}
		};
		
		OperationManager.getManager().addOperation(op, true);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result ICapture(final UUID sessionId) throws InterruptedException {
		trace("SensorService.ICapture(UUID)");
		
		Operation op = new Operation(OperationPurpose.CAPTURE) {
			@Override
			public Result execute() throws InterruptedException {
				//return capture(sessionId);
				Result result = null;
				result = capture(sessionId);
				
				return result;
			}
		};

		OperationManager.getManager().addOperation(op, true);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IDownload(final UUID captureId) {
		trace("SensorService.IDownload(UUID)");
		
		Operation op = new Operation(OperationPurpose.DOWNLOAD) {
			@Override
			public Result execute() throws InterruptedException {
				return download(captureId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IGetDownloadInformation(final UUID captureId) throws InterruptedException {
		trace("SensorService.IGetDownloadInformation(UUID)");
		
		Operation op = new Operation(OperationPurpose.GET_DOWNLOAD_INFORMATION) {
			@Override
			public Result execute() throws InterruptedException {
				return getDownloadInformation(captureId);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}

	@Override
	public Result IThriftyDownload(final UUID captureId, final String maxSize) throws InterruptedException {
		trace("SensorService.IThriftyDownload(UUID, String)");
		
		Operation op = new Operation(OperationPurpose.THRIFTY_DOWNLOAD) {
			@Override
			public Result execute() throws InterruptedException {
				return thriftyDownload(captureId, maxSize);
			}
		};
		
		OperationManager.getManager().addOperation(op, false);
		Result result = op.getResult();
		
		return result;
	}
	

	//TODO STREAMING!!!
	@Override
	public InputStream MultipartXMixedReplace() {
		trace("SensorService.IStream()");
		
//		Operation op = new Operation(OperationPurpose.STREAM){
//			@Override
//			public Result execute() throws InterruptedException {
//				return null;
//			}
//		};
//		
//		OperationManager.getManager().addOperation(op, false);
//		Result result = op.getResult();
		
		return stream();
	}
	
	
	
	
	/*@GET
	@Produces("multipart/x-mixed-replace")
	@Path("Goodbye")
	public Response sayGoodbye() {
		String boundary = UUID.randomUUID().toString();
		
		TestStream t = new TestStream(boundary);
		
		return Response.ok().entity(t).header("Connection", "keep-alive").type(String.format("multipart/x-mixed-replace; boundary=%s", boundary)).build();
	}
	*/
}